import json
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

from requests.exceptions import HTTPError

from pymatillion.constants import (
    BASE_URL,
    PASSWORD,
    PROJECT_GROUP_NAME,
    PROJECT_NAME,
    USERNAME,
)
from pymatillion.matillion import MatillionClient

DUMMY_URL = "https://fakematillioninstance.com"
DUMMY_USERNAME = "fake_user"
DUMMY_PASSWORD = "fake_pass"
DUMMY_PROJECT_GROUP_NAME = "Sample_Project_Group"
DUMMY_PROJECT_NAME = "Sample_Project_1"
DUMMY_JOB_NAME = "Sample_Job_1"
DUMMY_TASK_ID = 12345


PARENT_DIR = Path(__file__).parent


class TestMatillionClient(TestCase):
    client: MatillionClient

    def setUp(self):
        self.client = MatillionClient(DUMMY_URL, DUMMY_USERNAME, DUMMY_PASSWORD)

    def test_attributes(self):
        self.client.project_name = DUMMY_PROJECT_NAME
        self.client.project_group_name = DUMMY_PROJECT_GROUP_NAME

        self.assertEqual(self.client.base_url, DUMMY_URL)
        self.assertEqual(self.client.username, DUMMY_USERNAME)
        self.assertEqual(self.client.password, DUMMY_PASSWORD)
        self.assertEqual(self.client.project_group_name, DUMMY_PROJECT_GROUP_NAME)
        self.assertEqual(self.client.project_name, DUMMY_PROJECT_NAME)

    @patch("pymatillion.matillion.MatillionClient._ensure_attributes")
    @patch("pymatillion.matillion.requests.get")
    def test_list_project_groups(self, mock_get, mock_ensure_attributes):
        with open(PARENT_DIR.joinpath("list_project_groups_response.json")) as response:
            json_content = json.load(response)
        mock_get.return_value.json.return_value = json_content
        project_groups = self.client.list_project_groups()
        self.assertListEqual(project_groups, json_content)
        mock_ensure_attributes.called_once_with(BASE_URL, USERNAME, PASSWORD)

    @patch("pymatillion.matillion.MatillionClient._ensure_attributes")
    @patch("pymatillion.matillion.requests.get")
    def test_list_project_group_without_url(self, mock_get, mock_ensure_attributes):
        self.client.base_url = ""
        mock_ensure_attributes.side_effect = [
            ValueError(f"Undefined attributes: {BASE_URL}")
        ]
        with self.assertRaises(ValueError) as cm:
            self.client.list_project_groups()
        self.assertEqual(f"Undefined attributes: {BASE_URL}", cm.exception.args[0])
        mock_ensure_attributes.called_once_with(BASE_URL, USERNAME, PASSWORD)
        mock_get.assert_not_called()

    @patch("pymatillion.matillion.MatillionClient._ensure_attributes")
    @patch("pymatillion.matillion.requests.get")
    def test_list_project_group_without_username(
        self, mock_get, mock_ensure_attributes
    ):
        self.client.username = ""
        mock_ensure_attributes.side_effect = [
            ValueError(f"Undefined attributes: {USERNAME}")
        ]
        with self.assertRaises(ValueError) as cm:
            self.client.list_project_groups()
        self.assertEqual(f"Undefined attributes: {USERNAME}", cm.exception.args[0])
        mock_ensure_attributes.called_once_with(BASE_URL, USERNAME, PASSWORD)
        mock_get.assert_not_called()

    @patch("pymatillion.matillion.MatillionClient._ensure_attributes")
    @patch("pymatillion.matillion.requests.get")
    def test_list_project_group_without_password(
        self, mock_get, mock_ensure_attributes
    ):
        self.client.password = ""
        mock_ensure_attributes.side_effect = [
            ValueError(f"Undefined attributes: {PASSWORD}")
        ]
        with self.assertRaises(ValueError) as cm:
            self.client.list_project_groups()
        self.assertEqual(f"Undefined attributes: {PASSWORD}", cm.exception.args[0])
        mock_ensure_attributes.called_once_with(BASE_URL, USERNAME, PASSWORD)
        mock_get.assert_not_called()

    @patch("pymatillion.matillion.requests.get")
    def test_list_project_groups_when_attributes_unauthorized(self, mock_get):
        self.client.password = "xyz"
        self.client.username = "abc"

        with open(PARENT_DIR.joinpath("unauthorized_response.json")) as response:
            json_content = json.load(response)
        mock_get.side_effect = HTTPError(json_content["msg"])

        with self.assertRaises(HTTPError) as cm:
            self.client.list_project_groups()

        self.assertEqual(json_content["msg"], cm.exception.args[0])

    @patch("pymatillion.matillion.MatillionClient._ensure_attributes")
    @patch("pymatillion.matillion.requests.get")
    def test_list_projects(self, mock_get, mock_ensure_attributes):
        self.client.project_group_name = DUMMY_PROJECT_GROUP_NAME

        with open(PARENT_DIR.joinpath("list_projects_response.json")) as response:
            json_content = json.load(response)
        mock_get.return_value.json.return_value = json_content

        projects = self.client.list_projects()

        self.assertListEqual(projects, json_content)
        mock_ensure_attributes.called_once_with(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME
        )

    @patch("pymatillion.matillion.MatillionClient._ensure_attributes")
    @patch("pymatillion.matillion.requests.get")
    def test_list_projects_without_url(self, mock_get, mock_ensure_attributes):
        self.client.project_group_name = DUMMY_PROJECT_GROUP_NAME
        self.client.base_url = ""

        mock_ensure_attributes.side_effect = [
            ValueError(f"Undefined attributes: {BASE_URL}")
        ]

        with self.assertRaises(ValueError) as cm:
            self.client.list_projects()

        self.assertEqual(f"Undefined attributes: {BASE_URL}", cm.exception.args[0])
        mock_ensure_attributes.called_once_with(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME
        )
        mock_get.assert_not_called()

    @patch("pymatillion.matillion.MatillionClient._ensure_attributes")
    @patch("pymatillion.matillion.requests.get")
    def test_list_projects_without_password(self, mock_get, mock_ensure_attributes):
        self.client.project_group_name = DUMMY_PROJECT_GROUP_NAME
        self.client.password = ""

        mock_ensure_attributes.side_effect = [
            ValueError(f"Undefined attributes: {PASSWORD}")
        ]

        with self.assertRaises(ValueError) as cm:
            self.client.list_projects()

        self.assertEqual(f"Undefined attributes: {PASSWORD}", cm.exception.args[0])
        mock_ensure_attributes.called_once_with(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME
        )
        mock_get.assert_not_called()

    @patch("pymatillion.matillion.requests.get")
    def test_list_projects_when_attributes_unauthorized(self, mock_get):
        self.client.project_group_name = DUMMY_PROJECT_GROUP_NAME

        self.client.password = "xyz"
        self.client.username = "abc"

        with open(PARENT_DIR.joinpath("unauthorized_response.json")) as response:
            json_content = json.load(response)
        mock_get.side_effect = HTTPError(json_content["msg"])

        with self.assertRaises(HTTPError) as cm:
            self.client.list_projects()

        self.assertEqual(json_content["msg"], cm.exception.args[0])

    @patch("pymatillion.matillion.MatillionClient._ensure_attributes")
    @patch("pymatillion.matillion.requests.get")
    def test_list_jobs(self, mock_get, mock_ensure_attributes):
        self.client.project_group_name = DUMMY_PROJECT_GROUP_NAME
        self.client.project_name = DUMMY_PROJECT_NAME

        with open(PARENT_DIR.joinpath("list_jobs_response.json")) as response:
            json_content = json.load(response)
        mock_get.return_value.json.return_value = json_content

        jobs = self.client.list_jobs()

        self.assertListEqual(jobs, json_content)
        mock_ensure_attributes.called_once_with(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME, PROJECT_NAME
        )

    @patch("pymatillion.matillion.MatillionClient._ensure_attributes")
    @patch("pymatillion.matillion.requests.get")
    def test_list_jobs_without_url(self, mock_get, mock_ensure_attributes):
        self.client.project_group_name = DUMMY_PROJECT_GROUP_NAME
        self.client.project_name = DUMMY_PROJECT_NAME
        self.client.base_url = ""

        mock_ensure_attributes.side_effect = [
            ValueError(f"Undefined attributes: {BASE_URL}")
        ]

        with self.assertRaises(ValueError) as cm:
            self.client.list_jobs()

        self.assertEqual(f"Undefined attributes: {BASE_URL}", cm.exception.args[0])
        mock_ensure_attributes.called_once_with(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME, PROJECT_NAME
        )
        mock_get.assert_not_called()

    @patch("pymatillion.matillion.MatillionClient._ensure_attributes")
    @patch("pymatillion.matillion.requests.get")
    def test_list_jobs_without_password(self, mock_get, mock_ensure_attributes):
        self.client.project_group_name = DUMMY_PROJECT_GROUP_NAME
        self.client.project_name = DUMMY_PROJECT_NAME
        self.client.password = ""

        mock_ensure_attributes.side_effect = [
            ValueError(f"Undefined attributes: {PASSWORD}")
        ]

        with self.assertRaises(ValueError) as cm:
            self.client.list_jobs()

        self.assertEqual(f"Undefined attributes: {PASSWORD}", cm.exception.args[0])
        mock_ensure_attributes.called_once_with(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME, PROJECT_NAME
        )
        mock_get.assert_not_called()

    @patch("pymatillion.matillion.requests.get")
    def test_list_jobs_when_attributes_unauthorized(self, mock_get):
        self.client.project_group_name = DUMMY_PROJECT_GROUP_NAME
        self.client.project_name = DUMMY_PROJECT_NAME

        self.client.password = "xyz"
        self.client.username = "abc"

        with open(PARENT_DIR.joinpath("unauthorized_response.json")) as response:
            json_content = json.load(response)
        mock_get.side_effect = HTTPError(json_content["msg"])

        with self.assertRaises(HTTPError) as cm:
            self.client.list_jobs()

        self.assertEqual(json_content["msg"], cm.exception.args[0])

    @patch("pymatillion.matillion.MatillionClient._ensure_attributes")
    @patch("pymatillion.matillion.requests.post")
    def test_run_job(self, mock_post, mock_ensure_attributes):
        self.client.project_group_name = DUMMY_PROJECT_GROUP_NAME
        self.client.project_name = DUMMY_PROJECT_NAME

        with open(PARENT_DIR.joinpath("run_job_response.json")) as response:
            json_content = json.load(response)
        mock_post.return_value.json.return_value = json_content

        run_job_response = self.client.run_job(job_name=DUMMY_JOB_NAME)

        self.assertEqual(run_job_response, json_content)
        mock_ensure_attributes.called_once_with(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME, PROJECT_NAME
        )

    @patch("pymatillion.matillion.MatillionClient._ensure_attributes")
    @patch("pymatillion.matillion.requests.get")
    def test_get_task_details(self, mock_get, mock_ensure_attributes):
        self.client.project_group_name = DUMMY_PROJECT_GROUP_NAME
        self.client.project_name = DUMMY_PROJECT_NAME

        with open(PARENT_DIR.joinpath("task_detail_response.json")) as response:
            json_content = json.load(response)
        mock_get.return_value.json.return_value = json_content

        task_detail_response = self.client.get_task_details(task_id=DUMMY_TASK_ID)

        self.assertEqual(task_detail_response, json_content)
        mock_ensure_attributes.called_once_with(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME, PROJECT_NAME
        )

    @patch("pymatillion.matillion.MatillionClient._ensure_attributes")
    @patch("pymatillion.matillion.requests.post")
    def test_delete_project(self, mock_post, mock_ensure_attributes):
        self.client.project_group_name = DUMMY_PROJECT_GROUP_NAME
        self.client.project_name = DUMMY_PROJECT_NAME

        with open(PARENT_DIR.joinpath("delete_project_response.json")) as response:
            json_content = json.load(response)
        mock_post.return_value.json.return_value = json_content

        delete_project_response = self.client.delete_project(
            project_name=DUMMY_PROJECT_NAME
        )

        self.assertEqual(delete_project_response, json_content)
        mock_ensure_attributes.called_once_with(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME, PROJECT_NAME
        )

    @patch("pymatillion.matillion.MatillionClient._ensure_attributes")
    @patch("pymatillion.matillion.requests.post")
    def test_delete_job(self, mock_post, mock_ensure_attributes):
        self.client.project_group_name = DUMMY_PROJECT_GROUP_NAME
        self.client.project_name = DUMMY_PROJECT_NAME

        with open(PARENT_DIR.joinpath("delete_job_response.json")) as response:
            json_content = json.load(response)
        mock_post.return_value.json.return_value = json_content

        delete_job_response = self.client.delete_job(job_name=DUMMY_JOB_NAME)

        self.assertEqual(delete_job_response, json_content)
        mock_ensure_attributes.called_once_with(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME, PROJECT_NAME
        )



import logging
from typing import Dict, List

import requests

from pymatillion.constants import (
    API_GET,
    API_POST,
    BASE_URL,
    DEFAULT_VERSION,
    PASSWORD,
    PROJECT_GROUP_NAME,
    PROJECT_NAME,
    USERNAME,
)

logger = logging.getLogger(__name__)


class MatillionClient:
    def __init__(
        self, base_url, username, password, project_group_name=None, project_name=None
    ):
        self.base_url = base_url
        self.username = username
        self.password = password
        self.project_group_name = project_group_name
        self.project_name = project_name
        self._valid_attributes = False
        self._headers = None

    def _ensure_attributes(self, *args):
        attributes_to_check = set(args)
        invalid_attributes = []
        for attr in attributes_to_check:
            if (getattr(self, attr) is None) or (not getattr(self, attr)):
                invalid_attributes.append(attr)
        if invalid_attributes:
            raise ValueError(f'Undefined attributes: {",".join(invalid_attributes)}')
        else:
            self._valid_attributes = True

    def _set_headers(self):
        self._headers = {"Content-Type": "application/json"}

    def _api_request(
        self, http_method, *args, is_json=True, **kwargs
    ) -> requests.Response:
        """
        Make an API Request.

        Args:
            http_method (str) : HTTP method to use when making API request.
            args : Variable length list of strings to construct endpoint path.
            is_json (bool) : Set to False if payload/response is not JSON data.
            kwargs : Arbitrary keyword arguments to construct request payload.

        Returns:
            requests.Response : Response object.
        """

        api_request = getattr(requests, http_method)
        api_path = f'{self.base_url}/rest/v1/{"/".join(args)}'
        if is_json:
            if len(kwargs) == 1 and "json" in kwargs:
                response = api_request(
                    api_path,
                    headers=self._headers,
                    auth=(self.username, self.password),
                    json=kwargs["json"],
                )
            else:
                response = api_request(
                    api_path,
                    headers=self._headers,
                    auth=(self.username, self.password),
                    json=kwargs,
                )
        else:
            if len(kwargs) == 1 and "data" in kwargs:
                response = api_request(
                    api_path,
                    headers=self._headers,
                    auth=(self.username, self.password),
                    data=kwargs["data"],
                )
            else:
                response = api_request(
                    api_path,
                    headers=self._headers,
                    auth=(self.username, self.password),
                    data=kwargs,
                )
        try:
            response.raise_for_status()
        except Exception:
            logger.error(f"Response content:\n{response.content}")
            raise
        return response

    def list_project_groups(self) -> List[str]:
        """
        Retrieve Project Groups within the Matillion instance.

        Returns:
            Project Groups (list) : A list of strings with names of Project Groups.
        """
        self._ensure_attributes(BASE_URL, USERNAME, PASSWORD)
        return self._api_request(API_GET, "group").json()

    def list_projects(self) -> List[str]:
        """
        Retreive Projects within the specified Project Group of Matillion instance.

        Returns:
            Project Names (list) : A list of strings with names of Projects.
        """
        self._ensure_attributes(BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME)
        return self._api_request(
            API_GET, "group", "name", self.project_group_name, "project"
        ).json()

    def list_jobs(
        self, project_name: str = None, version: str = DEFAULT_VERSION
    ) -> List[str]:
        """
        Retrieve jobs within the specified version of the Matillion project.

        Args:
            project_name (str): Name of the Matillion project.
            version: (str), optional. Default value : default.
        Returns:
            Matillion Jobs (list): A list of of strings with names of the Matillion jobs.
        """
        self._ensure_attributes(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME, PROJECT_NAME
        )
        return self._api_request(
            API_GET,
            "group",
            "name",
            self.project_group_name,
            "project",
            "name",
            project_name if project_name else self.project_name,
            "version",
            "name",
            version,
            "job",
        ).json()

    def run_job(
        self,
        job_name: str,
        job_variables: dict = {},
        grid_variables: dict = {},
        project_name: str = None,
        version: str = DEFAULT_VERSION,
    ) -> Dict:
        """
        Run Matillion job within the specified project, for the specified version.

        Args:
            job_name (str): Name of the Matillion job intended to run.
            job_variables (dict): Dictionary of Matillion job variables.
            grid_variables (dict): Dictionary of Matillion grid variables.
            project_name (str): Name of the Matillion project.
            version: (str), optional. Default value : default.
        Returns:
            dict : Sample response can be found at https://documentation.matillion.com/docs/2475544#server-response
        """
        self._ensure_attributes(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME, PROJECT_NAME
        )
        body = {"scalarVariables": job_variables, "gridVariables": grid_variables}
        return self._api_request(
            API_POST,
            "group",
            "name",
            self.project_group_name,
            "project",
            "name",
            project_name if project_name else self.project_name,
            "version",
            "name",
            version,
            "job",
            "name",
            job_name,
            "run",
            json=body,
        ).json()

    def get_task_details(self, task_id: int, project_name: str = None) -> Dict:
        """
        Retrieve details for the specified executed task id.
        Args:
            task_id (int): ID of the executed Matillion task.
            project_name (str): Name of the Matillion project.
        Returns:
            dict : Sample response can be found at https://documentation.matillion.com/docs/2972278
        """
        self._ensure_attributes(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME, PROJECT_NAME
        )
        return self._api_request(
            API_GET,
            "group",
            "name",
            self.project_group_name,
            "project",
            "name",
            project_name if project_name else self.project_name,
            "task",
            "id",
            str(task_id),
        ).json()

    def delete_project(self, project_name: str, version: str = DEFAULT_VERSION) -> Dict:
        """
        Delete Matillion project. If version is specified then delete specified project version.

        Args:
            project_name (str): Name of the Matillion project.
            version: (str), optional. Default value : default.
        Returns:
            dict : Sample response can be found at https://documentation.matillion.com/docs/2949951#deleting-resources
        """
        self._ensure_attributes(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME, PROJECT_NAME
        )
        return self._api_request(
            API_POST,
            "group",
            "name",
            self.project_group_name,
            "project",
            "name",
            project_name if project_name else self.project_name,
            "version",
            "name",
            version,
            "delete",
        ).json()

    def delete_job(
        self, job_name: str, project_name: str = None, version: str = DEFAULT_VERSION
    ) -> Dict:
        """
        Delete Matillion job within the specified version of the project.

        Args:
            job_name: (str), required. Name of the Matillion job to delete.
            project_name: (str), optional. Name of the Matillion project.
            version: (str), optional. Default value : default.
        Returns:
            Dict[Any, Any] : Sample response can be found at https://documentation.matillion.com/docs/2949951#deleting-resources
        """
        self._ensure_attributes(
            BASE_URL, USERNAME, PASSWORD, PROJECT_GROUP_NAME, PROJECT_NAME
        )
        return self._api_request(
            API_POST,
            "group",
            "name",
            self.project_group_name,
            "project",
            "name",
            project_name if project_name else self.project_name,
            "version",
            "name",
            version,
            "job",
            "name",
            job_name,
            "delete",
        ).json()



BASE_URL = "base_url"
USERNAME = "username"
PASSWORD = "password"
PROJECT_GROUP_NAME = "project_group_name"
PROJECT_NAME = "project_name"
PROJECT_VERSION = "project_version"
DEFAULT_VERSION = "default"

# API HTTP Request Methods
API_DELETE = "delete"
API_GET = "get"
API_POST = "post"
API_PUT = "put"
